@@ -1,21 +1,16 @@
 import hashlib
-import sys
-import os
-import traceback
 import pdb
+import datetime
 
 from logzero import logger
-
 from neo.VM.RandomAccessStack import RandomAccessStack
 from neo.VM.ExecutionContext import ExecutionContext
 from neo.VM import VMState
 from neo.VM.OpCode import *
-from neo.SmartContract.ContractParameterType import ContractParameterType
-from neocore.BigInteger import BigInteger
-from neo.VM.InteropService import Array, Struct, StackItem
+from neo.VM.InteropService import Array, Struct, StackItem, CollectionMixin, Map, Boolean
 from neocore.UInt160 import UInt160
-import datetime
 from neo.Settings import settings
+from neo.VM.VMFault import VMFault
 
 
 class ExecutionEngine():
@@ -136,7 +131,7 @@ class ExecutionEngine():
         astack = self._AltStack
 
         if opcode > PUSH16 and opcode != RET and context.PushOnly:
-            self._VMState |= VMState.FAULT
+            return self.VM_FAULT_and_report(VMFault.UNKNOWN1)
 
         if opcode >= PUSHBYTES1 and opcode <= PUSHBYTES75:
             bytestoread = context.OpReader.ReadBytes(int.from_bytes(opcode, 'little'))
@@ -170,8 +165,7 @@ class ExecutionEngine():
                 offset = context.InstructionPointer + offset_b - 3
 
                 if offset < 0 or offset > len(context.Script):
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.INVALID_JUMP)
 
                 fValue = True
                 if opcode > JMP:
@@ -194,8 +188,7 @@ class ExecutionEngine():
 
             elif opcode == APPCALL or opcode == TAILCALL:
                 if self._Table is None:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN2)
 
                 script_hash = context.OpReader.ReadBytes(20)
 
@@ -211,8 +204,7 @@ class ExecutionEngine():
 
                 if script is None:
                     logger.error("Could not find script from script table: %s " % script_hash)
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.INVALID_CONTRACT, script_hash)
 
                 if opcode == TAILCALL:
                     istack.Pop().Dispose()
@@ -223,7 +215,7 @@ class ExecutionEngine():
                 call = context.OpReader.ReadVarBytes(252).decode('ascii')
                 self.write_log(call)
                 if not self._Service.Invoke(call, self):
-                    self._VMState |= VMState.FAULT
+                    return self.VM_FAULT_and_report(VMFault.SYSCALL_ERROR, call)
 
             # stack operations
             elif opcode == DUPFROMALTSTACK:
@@ -246,8 +238,7 @@ class ExecutionEngine():
                 n = estack.Pop().GetBigInteger()
 
                 if n < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN3)
 
                 # if n == 0 break, same as do x if n > 0
                 if n > 0:
@@ -259,8 +250,7 @@ class ExecutionEngine():
                 n = estack.Pop().GetBigInteger()
 
                 if n <= 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN4)
 
                 estack.Insert(n, estack.Peek())
 
@@ -289,8 +279,7 @@ class ExecutionEngine():
 
                 n = estack.Pop().GetBigInteger()
                 if n < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN5)
 
                 estack.PushT(estack.Peek(n))
 
@@ -298,8 +287,7 @@ class ExecutionEngine():
 
                 n = estack.Pop().GetBigInteger()
                 if n < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN6)
 
                 if n > 0:
                     estack.PushT(estack.Remove(n))
@@ -338,13 +326,11 @@ class ExecutionEngine():
 
                 count = estack.Pop().GetBigInteger()
                 if count < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.SUBSTR_INVALID_LENGTH)
 
                 index = estack.Pop().GetBigInteger()
                 if index < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.SUBSTR_INVALID_INDEX)
 
                 x = estack.Pop().GetByteArray()
 
@@ -354,8 +340,7 @@ class ExecutionEngine():
 
                 count = estack.Pop().GetBigInteger()
                 if count < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.LEFT_INVALID_COUNT)
 
                 x = estack.Pop().GetByteArray()
                 estack.PushT(x[:count])
@@ -364,13 +349,11 @@ class ExecutionEngine():
 
                 count = estack.Pop().GetBigInteger()
                 if count < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.RIGHT_INVALID_COUNT)
 
                 x = estack.Pop().GetByteArray()
                 if len(x) < count:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.RIGHT_UNKNOWN)
 
                 estack.PushT(x[-count:])
 
@@ -602,14 +585,14 @@ class ExecutionEngine():
                 except Exception as e:
                     estack.PushT(False)
                     logger.error("Could not checksig: %s " % e)
+                return
 
             elif opcode == CHECKMULTISIG:
 
                 n = estack.Pop().GetBigInteger()
 
                 if n < 1:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.CHECKMULTISIG_INVALID_PUBLICKEY_COUNT)
 
                 pubkeys = []
                 for i in range(0, n):
@@ -618,8 +601,7 @@ class ExecutionEngine():
                 m = estack.Pop().GetBigInteger()
 
                 if m < 1 or m > n:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.CHECKMULTISIG_SIGNATURE_ERROR, m, n)
 
                 sigs = []
 
@@ -655,22 +637,20 @@ class ExecutionEngine():
                 item = estack.Pop()
 
                 if not item:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN7)
 
-                if not item.IsArray:
-                    estack.PushT(len(item.GetByteArray()))
+                if isinstance(item, CollectionMixin):
+                    estack.PushT(item.Count)
 
                 else:
-                    estack.PushT(len(item.GetArray()))
+                    estack.PushT(len(item.GetByteArray()))
 
             elif opcode == PACK:
 
                 size = estack.Pop().GetBigInteger()
 
                 if size < 0 or size > estack.Count:
-                    self._VMState |= VMState.FAULT
-                    return
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN8)
 
                 items = []
 
@@ -683,9 +663,8 @@ class ExecutionEngine():
             elif opcode == UNPACK:
                 item = estack.Pop()
 
-                if not item.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
+                if not isinstance(item, Array):
+                    return self.VM_FAULT_and_report(VMFault.UNPACK_INVALID_TYPE, item)
 
                 items = item.GetArray()
                 items.reverse()
@@ -696,74 +675,92 @@ class ExecutionEngine():
 
             elif opcode == PICKITEM:
 
-                index = estack.Pop().GetBigInteger()
-                if index < 0:
-                    self._VMState |= VMState.FAULT
-                    return
+                key = estack.Pop()
 
-                item = estack.Pop()
+                if isinstance(key, CollectionMixin):
+                    # key must be an array index or dictionary key, but not a collection
+                    return self.VM_FAULT_and_report(VMFault.KEY_IS_COLLECTION, key)
 
-                if not item.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
+                collection = estack.Pop()
 
-                items = item.GetArray()
+                if isinstance(collection, Array):
+                    index = key.GetBigInteger()
+                    if index < 0 or index >= collection.Count:
+                        return self.VM_FAULT_and_report(VMFault.PICKITEM_INVALID_INDEX, index, collection.Count)
 
-                if index >= len(items):
-                    self._VMState |= VMState.FAULT
-                    return
+                    items = collection.GetArray()
+                    to_pick = items[index]
+                    estack.PushT(to_pick)
 
-                to_pick = items[index]
+                elif isinstance(collection, Map):
+                    success, value = collection.TryGetValue(key)
 
-                estack.PushT(to_pick)
+                    if success:
+                        estack.PushT(value)
+                    else:
+                        return self.VM_FAULT_and_report(VMFault.DICT_KEY_NOT_FOUND, key, collection.Keys)
+
+                else:
+                    return self.VM_FAULT_and_report(VMFault.PICKITEM_INVALID_TYPE, key, collection)
 
             elif opcode == SETITEM:
-                newItem = estack.Pop()
+                value = estack.Pop()
 
-                if issubclass(type(newItem), StackItem) and newItem.IsStruct:
-                    newItem = newItem.Clone()
+                if isinstance(value, Struct):
+                    value = value.Clone()
 
-                index = estack.Pop().GetBigInteger()
+                key = estack.Pop()
 
-                arrItem = estack.Pop()
+                if isinstance(key, CollectionMixin):
+                    return self.VM_FAULT_and_report(VMFault.KEY_IS_COLLECTION)
 
-                if not issubclass(type(arrItem), StackItem) or not arrItem.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
+                collection = estack.Pop()
 
-                items = arrItem.GetArray()
+                if isinstance(collection, Array):
 
-                if index < 0 or index >= len(items):
-                    self._VMState |= VMState.FAULT
-                    return
+                    index = key.GetBigInteger()
 
-                items[index] = newItem
+                    if index < 0 or index >= collection.Count:
+                        return self.VM_FAULT_and_report(VMFault.SETITEM_INVALID_INDEX)
+
+                    items = collection.GetArray()
+                    items[index] = value
+
+                elif isinstance(collection, Map):
+
+                    collection.SetItem(key, value)
+
+                else:
+
+                    return self.VM_FAULT_and_report(VMFault.SETITEM_INVALID_TYPE, key, collection)
 
             elif opcode == NEWARRAY:
 
                 count = estack.Pop().GetBigInteger()
-                items = [None for i in range(0, count)]
+                items = [Boolean(False) for i in range(0, count)]
                 estack.PushT(Array(items))
 
             elif opcode == NEWSTRUCT:
 
                 count = estack.Pop().GetBigInteger()
 
-                items = [None for i in range(0, count)]
+                items = [Boolean(False) for i in range(0, count)]
 
                 estack.PushT(Struct(items))
 
+            elif opcode == NEWMAP:
+                estack.PushT(Map())
+
             elif opcode == APPEND:
                 newItem = estack.Pop()
 
-                if type(newItem) is Struct:
+                if isinstance(newItem, Struct):
                     newItem = newItem.Clone()
 
                 arrItem = estack.Pop()
 
-                if not arrItem.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
+                if not isinstance(arrItem, Array):
+                    return self.VM_FAULT_and_report(VMFault.APPEND_INVALID_TYPE, arrItem)
 
                 arr = arrItem.GetArray()
                 arr.append(newItem)
@@ -771,41 +768,105 @@ class ExecutionEngine():
             elif opcode == REVERSE:
 
                 arrItem = estack.Pop()
-                if not arrItem.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
-                arrItem.GetArray().reverse()
+                if not isinstance(arrItem, Array):
+                    return self.VM_FAULT_and_report(VMFault.REVERSE_INVALID_TYPE, arrItem)
+
+                arrItem.Reverse()
 
             elif opcode == REMOVE:
-                index = estack.Pop().GetBigInteger()
-                arrItem = estack.Pop()
-                if not arrItem.IsArray:
-                    self._VMState |= VMState.FAULT
-                    return
-                items = arrItem.GetArray()
 
-                if index < 0 or index >= len(items):
-                    self._VMState |= VMState.FAULT
-                    return
-                del items[index]
+                key = estack.Pop()
+
+                if isinstance(key, CollectionMixin):
+                    return self.VM_FAULT_and_report(VMFault.UNKNOWN1)
+
+                collection = estack.Pop()
+
+                if isinstance(collection, Array):
+
+                    index = key.GetBigInteger()
+
+                    if index < 0 or index >= collection.Count:
+                        return self.VM_FAULT_and_report(VMFault.REMOVE_INVALID_INDEX, index, collection.Count)
+
+                    collection.RemoveAt(index)
+
+                elif isinstance(collection, Map):
+
+                    collection.Remove(key)
+
+                else:
+
+                    return self.VM_FAULT_and_report(VMFault.REMOVE_INVALID_TYPE, key, collection)
+
+            elif opcode == HASKEY:
+
+                key = estack.Pop()
+
+                if isinstance(key, CollectionMixin):
+                    return self.VM_FAULT_and_report(VMFault.DICT_KEY_ERROR)
+
+                collection = estack.Pop()
+
+                if isinstance(collection, Array):
+
+                    index = key.GetBigInteger()
+
+                    if index < 0:
+                        return self.VM_FAULT_and_report(VMFault.DICT_KEY_ERROR)
+
+                    estack.PushT(index < collection.Count)
+
+                elif isinstance(collection, Map):
+
+                    estack.PushT(collection.ContainsKey(key))
+
+                else:
+
+                    return self.VM_FAULT_and_report(VMFault.DICT_KEY_ERROR)
+
+            elif opcode == KEYS:
+
+                collection = estack.Pop()
+
+                if isinstance(collection, Map):
+
+                    estack.PushT(Array(collection.Keys))
+                else:
+                    return self.VM_FAULT_and_report(VMFault.DICT_KEY_ERROR)
+
+            elif opcode == VALUES:
+
+                collection = estack.Pop()
+                values = []
+
+                if isinstance(collection, Map):
+                    values = collection.Values
+
+                elif isinstance(collection, Array):
+                    values = collection
+
+                else:
+                    return self.VM_FAULT_and_report(VMFault.DICT_KEY_ERROR)
+
+                newArray = Array()
+                for item in values:
+                    if isinstance(item, Struct):
+                        newArray.Add(item.Clone())
+                    else:
+                        newArray.Add(item)
+
+                estack.PushT(newArray)
 
             elif opcode == THROW:
-                self._VMState |= VMState.FAULT
-                return
+                return self.VM_FAULT_and_report(VMFault.THROW)
 
             elif opcode == THROWIFNOT:
                 if not estack.Pop().GetBoolean():
-                    self._VMState |= VMState.FAULT
-                    return
-
-            elif opcode == DEBUG:
-                pdb.set_trace()
-                return
+                    return self.VM_FAULT_and_report(VMFault.THROWIFNOT)
 
             else:
-
-                self._VMState |= VMState.FAULT
-                return
+                return self.VM_FAULT_and_report(VMFault.UNKNOWN_OPCODE, opcode)
 
         if self._VMState & VMState.FAULT == 0 and self.InvocationStack.Count > 0:
 
@@ -841,10 +902,10 @@ class ExecutionEngine():
         else:
             op = self.CurrentContext.OpReader.ReadByte(do_ord=False)
 
-        #        opname = ToName(op)
-        #        logger.info("____________________________________________________")
-        #        logger.info("[%s] [%s] %02x -> %s" % (self.CurrentContext.InstructionPointer,self.ops_processed,int.from_bytes(op,byteorder='little'), opname))
-        #        logger.info("-----------------------------------")
+#        opname = ToName(op)
+#        logger.info("____________________________________________________")
+#        logger.info("[%s] [%s] %02x -> %s" % (self.CurrentContext.InstructionPointer,self.ops_processed,int.from_bytes(op,byteorder='little'), opname))
+#        logger.info("-----------------------------------")
 
         self.ops_processed += 1
 
@@ -883,3 +944,82 @@ class ExecutionEngine():
                 self._VMState & VMState.BREAK == 0 and \
                 self._InvocationStack.Count > count:
             self.StepInto()
+
+    def VM_FAULT_and_report(self, id, *args):
+        self._VMState |= VMState.FAULT
+
+        if id == VMFault.INVALID_JUMP:
+            error_msg = "Attemping to JMP/JMPIF/JMPIFNOT to an invalid location."
+
+        elif id == VMFault.INVALID_CONTRACT:
+            script_hash = args[0]
+            error_msg = "Trying to call an unknown contract with script_hash {}\nMake sure the contract exists on the blockchain".format(script_hash)
+
+        elif id == VMFault.CHECKMULTISIG_INVALID_PUBLICKEY_COUNT:
+            error_msg = "CHECKMULTISIG - provided public key count is less than 1."
+
+        elif id == VMFault.CHECKMULTISIG_SIGNATURE_ERROR:
+            if args[0] < 1:
+                error_msg = "CHECKMULTISIG - Minimum required signature count cannot be less than 1."
+            else:  # m > n
+                m = args[0]
+                n = args[1]
+                error_msg = "CHECKMULTISIG - Insufficient signatures provided ({}). Minimum required is {}".format(m, n)
+
+        elif id == VMFault.UNPACK_INVALID_TYPE:
+            item = args[0]
+            error_msg = "Failed to UNPACK item. Item is not an array but of type: {}".format(type(item))
+
+        elif id == VMFault.PICKITEM_INVALID_TYPE:
+            index = args[0]
+            item = args[1]
+            error_msg = "Cannot access item at index {}. Item is not an array or dict but of type: {}".format(index, type(item))
+
+        elif id == VMFault.PICKITEM_NEGATIVE_INDEX:
+            error_msg = "Attempting to access an array using a negative index"
+
+        elif id == VMFault.PICKITEM_INVALID_INDEX:
+            index = args[0]
+            length = args[1]
+            error_msg = "Array index is less than zero or {} exceeds list length {}".format(index, length)
+
+        elif id == VMFault.APPEND_INVALID_TYPE:
+            item = args[0]
+            error_msg = "Cannot append to item. Item is not an array but of type: {}".format(type(item))
+
+        elif id == VMFault.REVERSE_INVALID_TYPE:
+            item = args[0]
+            error_msg = "Cannot REVERSE item. Item is not an array but of type: {}".format(type(item))
+
+        elif id == VMFault.REMOVE_INVALID_TYPE:
+            item = args[0]
+            index = args[1]
+            error_msg = "Cannot REMOVE item at index {}. Item is not an array but of type: {}".format(index, type(item))
+
+        elif id == VMFault.REMOVE_INVALID_INDEX:
+            index = args[0]
+            length = args[1]
+
+            if index < 0:
+                error_msg = "Cannot REMOVE item at index {}. Index < 0".format(index)
+
+            else:  # index >= len(items):
+                error_msg = "Cannot REMOVE item at index {}. Index exceeds array length {}".format(index, length)
+
+        elif id == VMFault.POP_ITEM_NOT_ARRAY:
+            error_msg = "Items(s) not array: %s" % [item for item in args]
+
+        elif id == VMFault.UNKNOWN_OPCODE:
+            opcode = args[0]
+            error_msg = "Unknown opcode found: {}".format(opcode)
+
+        else:
+            error_msg = id
+
+        # these get used a lot actually now, so we dont want them printed to the console
+        if id in [VMFault.THROW, VMFault.THROWIFNOT]:
+            logger.debug("({}) {}".format(self.ops_processed, id))
+        else:
+            logger.error("({}) {}".format(self.ops_processed, error_msg))
+
+        return
